from pandas.testing import assert_frame_equal

from akerbp.mlpet import feature_engineering
from akerbp.mlpet.tests.client import CLIENT, CLIENT_FUNCTIONS
from akerbp.mlpet.tests.data.data import DEPTH_TREND_X, DEPTH_TREND_Y
from akerbp.mlpet.tests.data.data import FORMATION_DF as FORMATION_DF_WITH_SYSTEMS
from akerbp.mlpet.tests.data.data import (
    FORMATION_TOPS_MAPPER,
    TEST_DF,
    VERTICAL_DEPTHS_MAPPER,
    VERTICAL_DF,
)
from akerbp.mlpet.tests.data.kubeflow_config import ID_COLUMN
from akerbp.mlpet.tests.data.kubeflow_config import (
    KUBEFLOW_MODEL_URL_VSH as KUBEFLOW_MODEL_URL,
)
from akerbp.mlpet.tests.data.kubeflow_config import VSH_HEADER, VSH_KWARGS

FORMATION_DF = FORMATION_DF_WITH_SYSTEMS.drop(columns=["SYSTEM"])
DEPTH_COL = "DEPTH"


def test_add_formations_and_groups_using_mapper():
    df_with_tops = feature_engineering.add_formations_and_groups(
        FORMATION_DF[[DEPTH_COL, ID_COLUMN]],
        formation_tops_mapper=FORMATION_TOPS_MAPPER,
        id_column=ID_COLUMN,
    )
    # Sorting columns because column order is not so important
    assert_frame_equal(df_with_tops.sort_index(axis=1), FORMATION_DF.sort_index(axis=1))


def test_add_formations_and_groups_using_client():
    df_with_tops = feature_engineering.add_formations_and_groups(
        FORMATION_DF[[DEPTH_COL, ID_COLUMN]],
        id_column=ID_COLUMN,
        client=CLIENT,
    )
    assert_frame_equal(df_with_tops.sort_index(axis=1), FORMATION_DF.sort_index(axis=1))


def test_add_formations_and_groups_using_client_with_systems():
    df_with_tops = feature_engineering.add_formations_and_groups(
        FORMATION_DF[[DEPTH_COL, ID_COLUMN]],
        id_column=ID_COLUMN,
        client=CLIENT,
        add_systems=True,
    )
    assert_frame_equal(
        df_with_tops.sort_index(axis=1), FORMATION_DF_WITH_SYSTEMS.sort_index(axis=1)
    )


def test_add_formations_and_groups_when_no_client_nor_mapping_is_provided():
    _ = feature_engineering.add_formations_and_groups(
        FORMATION_DF[[DEPTH_COL, ID_COLUMN]],
        id_column=ID_COLUMN,
    )


def test_add_vertical_depths_using_mapper():
    df_with_vertical_depths = feature_engineering.add_vertical_depths(
        VERTICAL_DF[[DEPTH_COL, ID_COLUMN]],
        vertical_depths_mapper=VERTICAL_DEPTHS_MAPPER,
        id_column=ID_COLUMN,
        md_column=DEPTH_COL,
    )

    assert_frame_equal(
        df_with_vertical_depths.sort_index(axis=1), VERTICAL_DF.sort_index(axis=1)
    )


def test_add_vertical_depths_using_client():
    df_with_vertical_depths = feature_engineering.add_vertical_depths(
        VERTICAL_DF[[DEPTH_COL, ID_COLUMN]],
        id_column=ID_COLUMN,
        md_column=DEPTH_COL,
        client=CLIENT,
    )

    assert_frame_equal(
        df_with_vertical_depths.sort_index(axis=1), VERTICAL_DF.sort_index(axis=1)
    )


def test_add_vertical_depths_when_no_client_nor_mapping_is_provided():
    _ = feature_engineering.add_vertical_depths(
        VERTICAL_DF[[DEPTH_COL, ID_COLUMN]],
        id_column=ID_COLUMN,
        md_column=DEPTH_COL,
    )


def test_add_well_metadata():
    metadata = {"30/11-6 S": {"FOO": 0}, "25/7-4 S": {"FOO": 1}}
    df = feature_engineering.add_well_metadata(
        TEST_DF,
        metadata_dict=metadata,
        metadata_columns=["FOO"],
        id_column=ID_COLUMN,
    )
    assert "FOO" in df.columns.tolist()


def test_add_depth_trend():
    result = feature_engineering.add_depth_trend(
        DEPTH_TREND_X,
        id_column="well",
        env="prod",
        return_file=False,
        return_CI=True,
        client=CLIENT_FUNCTIONS,
        keyword_arguments=dict(
            nan_numerical_value=-9999,
            nan_textual_value="MISSING",
        ),
    )

    assert DEPTH_TREND_Y.equals(result[DEPTH_TREND_Y.columns])


def test_add_petrophysical_features_add_VSH_call_CDF_model_return_only_vsh_aut():
    df = TEST_DF.rename(columns={"DENC": "DEN"})
    petrophysical_features = ["VSH"]
    result = feature_engineering.add_petrophysical_features(
        df=df,
        id_column=ID_COLUMN,
        petrophysical_features=petrophysical_features,
        keyword_arguments=VSH_KWARGS,
        client=CLIENT_FUNCTIONS,
    )
    output_curves = result.columns
    assert "VSH" in output_curves, "'VSH' not added to dataframe"


def test_add_petrophysical_features_add_VSH_from_kubeflow_return_only_vsh_aut():
    petrophysical_features = ["VSH"]
    result = feature_engineering.add_petrophysical_features(
        df=TEST_DF,
        id_column=ID_COLUMN,
        keyword_arguments=VSH_KWARGS,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        request_header=VSH_HEADER,
        petrophysical_features=petrophysical_features,
    )
    output_curves = set(result.columns)
    assert {"VSH"}.issubset(output_curves), "'VSH' not added to dataframe"


def test_add_petrophysical_features_add_VSH_from_kubeflow_return_composite_curves_no_CI():
    petrophysical_features = ["VSH"]
    keyword_arguments = VSH_KWARGS.copy()
    keyword_arguments["return_only_vsh_aut"] = False
    result = feature_engineering.add_petrophysical_features(
        df=TEST_DF,
        id_column=ID_COLUMN,
        keyword_arguments=keyword_arguments,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        request_header=VSH_HEADER,
        petrophysical_features=petrophysical_features,
    )
    output_curves = result.columns.tolist()
    assert {"VSH", "VSH_GR_AUT_QCFLAG"}.issubset(
        set(output_curves)
    ) and "VSH_AUT_P90" not in output_curves


def test_add_petrophysical_features_add_VSH_from_kubeflow_return_composite_curves_and_CI():
    petrophysical_features = ["VSH"]
    keyword_arguments = VSH_KWARGS.copy()
    keyword_arguments["return_only_vsh_aut"] = False
    result = feature_engineering.add_petrophysical_features(
        df=TEST_DF,
        id_column=ID_COLUMN,
        keyword_arguments=keyword_arguments,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        request_header=VSH_HEADER,
        return_CI=True,
        petrophysical_features=petrophysical_features,
    )
    output_curves = result.columns.tolist()
    assert {"VSH", "VSH_GR_AUT_QCFLAG", "VSH_AUT_P90"}.issubset(set(output_curves))



import numpy as np
import pandas as pd

from akerbp.mlpet import feature_engineering, petrophysical_features
from akerbp.mlpet.dataloader import DataLoader
from akerbp.mlpet.tests.client import CLIENT, CLIENT_FUNCTIONS
from akerbp.mlpet.tests.data.kubeflow_config import ID_COLUMN
from akerbp.mlpet.tests.data.kubeflow_config import (
    KUBEFLOW_MODEL_URL_VSH as KUBEFLOW_MODEL_URL,
)
from akerbp.mlpet.tests.data.kubeflow_config import VSH_HEADER, VSH_KWARGS

WELL = "15/3-5"


def test_guess_bs_from_cali():
    input = pd.DataFrame({"CALI": [6.1, 5.9, 12.0, 12.02]})
    df = petrophysical_features.guess_BS_from_CALI(input)
    assert "BS" in df.columns.tolist()


def test_calculate_cali_bs():
    input = pd.DataFrame({"CALI": np.array([6.1, 5.9, 12.0, 12.02])})
    df = petrophysical_features.calculate_CALI_BS(input)
    assert "CALI-BS" in df.columns.tolist()


def test_calculate_VSH():
    dl = DataLoader()
    df = dl.load_from_cdf(
        client=CLIENT, metadata={"wellbore_name": WELL, "subtype": "BEST"}
    )
    df[ID_COLUMN] = WELL
    df = petrophysical_features.calculate_LFI(df)
    # BS is required but missing in this sequence to init to all nans
    df["BS"] = np.nan
    df = feature_engineering.add_formations_and_groups(
        df, id_column=ID_COLUMN, depth_column="DEPTH"
    )
    df = feature_engineering.add_vertical_depths(
        df, id_column=ID_COLUMN, md_column="DEPTH"
    )

    df_out = petrophysical_features.calculate_VSH(
        df,
        id_column=ID_COLUMN,
        env="prod",
        return_CI=True,
        client=CLIENT_FUNCTIONS,
        keyword_arguments=dict(
            calculate_denneu=True,
            VSH_curves=["GR", "LFI"],
            groups_column_name="GROUP",
            formations_column_name="FORMATION",
            return_only_vsh_aut=False,
            nan_numerical_value=-9999,
            nan_textual_value="MISSING",
        ),
    )
    assert "VSH" in df_out.columns.tolist()


def test_calculate_VSH_call_kubeflow_model_return_vsh_aut_only_no_CI():
    dl = DataLoader()
    df = dl.load_from_cdf(
        client=CLIENT, metadata={"wellbore_name": WELL, "subtype": "BEST"}
    )
    df[ID_COLUMN] = WELL

    df_out = petrophysical_features.calculate_VSH(
        df=df,
        id_column=ID_COLUMN,
        keyword_arguments=VSH_KWARGS,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        request_header=VSH_HEADER,
    )
    output_columns = df_out.columns.tolist()
    assert "VSH" in output_columns and "VSH_AUT_P90" not in output_columns


def test_calculate_VSH_call_kubeflow_model_return_composite_curves_no_CI():
    dl = DataLoader()
    df = dl.load_from_cdf(
        client=CLIENT, metadata={"wellbore_name": WELL, "subtype": "BEST"}
    )
    df[ID_COLUMN] = WELL
    keyword_arguments = VSH_KWARGS.copy()
    keyword_arguments["return_only_vsh_aut"] = False

    df_out = petrophysical_features.calculate_VSH(
        df=df,
        id_column=ID_COLUMN,
        keyword_arguments=keyword_arguments,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        request_header=VSH_HEADER,
    )
    output_columns = df_out.columns.tolist()
    assert {"VSH", "VSH_GR_AUT_QCFLAG"}.issubset(
        set(output_columns)
    ) and "VSH_AUT_P90" not in output_columns


def test_calculate_VSH_call_kubeflow_model_return_composite_curves_and_CI():
    dl = DataLoader()
    df = dl.load_from_cdf(
        client=CLIENT, metadata={"wellbore_name": WELL, "subtype": "BEST"}
    )
    df[ID_COLUMN] = WELL
    keyword_arguments = VSH_KWARGS.copy()
    keyword_arguments["return_only_vsh_aut"] = False

    df_out = petrophysical_features.calculate_VSH(
        df=df,
        id_column=ID_COLUMN,
        keyword_arguments=keyword_arguments,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        request_header=VSH_HEADER,
        return_CI=True,
    )
    output_columns = df_out.columns.tolist()
    assert {"VSH", "VSH_GR_AUT_QCFLAG", "VSH_AUT_P90"}.issubset(set(output_columns))



import os

ID_COLUMN = "well_name"
KUBEFLOW_INGRESS_IP = os.environ.get("KUBEFLOW_INGRESS_IP")
KUBEFLOW_MODEL_URL_VSH = f"http://{KUBEFLOW_INGRESS_IP}/v1/models/automatic-vsh/predict"
API_KEY_VSH = os.environ.get("KUBEFLOW_API_KEY")
VSH_HEADER = {"KUBEFLOW_API_KEY": API_KEY_VSH}
VSH_KWARGS = {
    "nan_numerical_value": -9999,
    "nan_textual_value": "MISSING",
    "VSH_curves": ["GR"],
}



import os

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import APIKey

CLIENT_NAME = "MLPet testing"
CLIENT_NAME_FUNCTIONS = "MLPet testing, functions"
PROJECT = "akbp-subsurface"


class MissingEnvironmentVariableError(Exception):
    """Raised if an environment variable is missing"""


try:
    credentials = APIKey(os.environ["COGNITE_API_KEY"])
except KeyError as e:
    raise MissingEnvironmentVariableError(
        "Environment variable 'COGNITE_API_KEY' is missing, unable to initialize client for testing"
    ) from e

try:
    credentials_functions = APIKey(os.environ["COGNITE_API_KEY_FUNCTIONS"])
except KeyError as e:
    raise MissingEnvironmentVariableError(
        "Environment variable 'COGNITE_API_KEY_FUNCTIONS' is missing, unable to initialize client for testing"
    ) from e

client_config = ClientConfig(
    client_name=CLIENT_NAME,
    project=PROJECT,
    credentials=credentials,
)
client_config_functions = ClientConfig(
    client_name=CLIENT_NAME_FUNCTIONS,
    project=PROJECT,
    credentials=credentials_functions,
)

CLIENT = CogniteClient(config=client_config)
CLIENT_FUNCTIONS = CogniteClient(config=client_config_functions)



import os

import pandas as pd
import pytest
import yaml
from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import APIKey

from akerbp.mlpet import utilities as utils
from akerbp.mlpet.tests.data.data import (
    FORMATION_TOPS_MAPPER,
    TEST_DF,
    VERTICAL_DEPTHS_MAPPER,
)
from akerbp.mlpet.tests.data.kubeflow_config import ID_COLUMN, KUBEFLOW_INGRESS_IP
from akerbp.mlpet.tests.data.kubeflow_config import (
    KUBEFLOW_MODEL_URL_VSH as KUBEFLOW_MODEL_URL,
)
from akerbp.mlpet.tests.data.kubeflow_config import VSH_HEADER, VSH_KWARGS

# Assuming COGNITE_API_KEY is set in the environment
credentials = APIKey(os.environ["COGNITE_API_KEY"])
client_config = ClientConfig(
    client_name="test", project="akbp-subsurface", credentials=credentials
)
CLIENT = CogniteClient(client_config)
WELL_NAMES = ["25/10-10"]


def test_get_formation_tops():
    formation_tops_mapper = utils.get_formation_tops(WELL_NAMES, CLIENT)
    assert formation_tops_mapper == FORMATION_TOPS_MAPPER


def test_get_vertical_depths():
    retrieved_vertical_depths = utils.get_vertical_depths(WELL_NAMES, CLIENT)
    # empty_queries should be an empty list for the provided WELL_NAMES
    assert retrieved_vertical_depths == VERTICAL_DEPTHS_MAPPER


def test_remove_wo_label():
    df = utils.drop_rows_wo_label(TEST_DF[["AC", "BS"]], label_column="BS")
    assert df.shape[0] == 8


def test_standardize_names():
    mapper = yaml.load(
        open("src/akerbp/mlpet/tests/data/test_mappings.yaml", "r"),
        Loader=yaml.SafeLoader,
    )
    utils.standardize_names(TEST_DF.columns.tolist(), mapper=mapper)


def test_standardize_group_formation_name():
    assert utils.standardize_group_formation_name("ØRN") == "ORN"


def test_map_formation_group_system():
    tests = pd.Series(["UNDIFFERENTIATED", "FOO BAR", "NO FORMAL NAME 1", "HeGrE"])
    tests = tests.apply(utils.standardize_group_formation_name)
    assert utils.map_formation_group_system(tests, MissingValue=-9999) == (
        ("UNKNOWN FM", -9999, "UNKNOWN FM", -9999),
        ("UNKNOWN GP", -9999, "UNKNOWN GP", "HEGRE GP"),
        (-9999, -9999, -9999, "TRIASSIC SY"),
    )


def test_get_well_metadata():
    metadata = utils.get_well_metadata(client=CLIENT, well_names=WELL_NAMES)
    assert metadata[WELL_NAMES[0]]["CDF_wellName"] == WELL_NAMES[0]


def test_get_cognite_client_no_args_returns_logged_in_client():
    client = utils.get_cognite_client()
    logged_in = client.login.status().logged_in
    assert logged_in


def test_get_cognite_client_pass_api_key_returns_logged_in_client():
    api_key = os.environ["COGNITE_API_KEY"]
    client = utils.get_cognite_client(
        cognite_api_key=api_key,
    )
    logged_in = client.login.status().logged_in
    assert logged_in


def test_get_cognite_client_wrong_api_key_return_client_not_logged_in():
    api_key = "wrong_key_go_home"
    client = utils.get_cognite_client(cognite_api_key=api_key)
    logged_in = client.login.status().logged_in
    assert not logged_in


def test_run_kubeflow_model_automatic_vsh_return_metadata():
    metadata_url = f"http://{KUBEFLOW_INGRESS_IP}/v1/models/automatic-vsh/metadata"
    metadata = utils.run_kubeflow_model(
        in_data={"input": {}}, model_url=metadata_url, header=VSH_HEADER
    )
    expected_metadata_fields = [
        "keyword_arguments",
        "model_name",
        "optional_input",
        "optional_output_curves",
        "output_curves",
        "petrel_exposure",
        "required_input",
        "supports_external_retrieval",
    ]
    assert list(metadata.keys()) == expected_metadata_fields


def test_run_kubeflow_model_automatic_vsh_invalid_endpoint_raise_valueerror():
    url = f"http://{KUBEFLOW_INGRESS_IP}/v1/models/automatic-vsh/not_a_valid_endpoint"
    with pytest.raises(ValueError):
        utils.run_kubeflow_model(in_data={"input": {}}, model_url=url)


def test_run_kubeflow_model_automatic_vsh_return_only_vsh_aut():
    df = TEST_DF.rename(columns={"DENC": "DEN"}).fillna(
        VSH_KWARGS["nan_numerical_value"]
    )
    in_data = {
        "input": [
            {
                "well": df.well_name.iloc[0],
                "input_logs": df.to_dict(orient="list"),
                "keyword_arguments": VSH_KWARGS,
            }
        ]
    }
    output_keys = utils.run_kubeflow_model(
        in_data=in_data,
        model_url=KUBEFLOW_MODEL_URL,
        header=VSH_HEADER,
    )[0].keys()
    expected_output_keys = ["VSH_AUT", "well"]
    assert list(output_keys) == expected_output_keys


def test_run_kubeflow_model_automatic_vsh_return_additional_curves():
    df = TEST_DF.rename(columns={"DENC": "DEN"}).fillna(
        VSH_KWARGS["nan_numerical_value"]
    )
    keyword_arguments = VSH_KWARGS.copy()
    keyword_arguments["return_only_vsh_aut"] = False
    in_data = {
        "input": [
            {
                "well": df.well_name.iloc[0],
                "input_logs": df.to_dict(orient="list"),
                "keyword_arguments": keyword_arguments,
            }
        ]
    }
    output_keys = utils.run_kubeflow_model(
        in_data=in_data,
        model_url=KUBEFLOW_MODEL_URL,
        header=VSH_HEADER,
    )[0].keys()
    expected_output_keys = [
        "BS_CHANGE",
        "GR_sh",
        "GR_ss",
        "VSH_AUT",
        "VSH_AUT_P025",
        "VSH_AUT_P10",
        "VSH_AUT_P16",
        "VSH_AUT_P84",
        "VSH_AUT_P90",
        "VSH_AUT_P975",
        "VSH_GR_AUT",
        "VSH_GR_AUT_INTER_RATIO",
        "VSH_GR_AUT_QCFLAG",
        "well",
    ]
    assert list(output_keys) == expected_output_keys


def test_run_kubeflow_model_automatic_vsh_raise_exception_wrong_address():
    ingress_ip_wrong = "1.2.3"
    url = f"http://{ingress_ip_wrong}/v1/models/automatic-vsh/metadata"
    with pytest.raises(ValueError):
        utils.run_kubeflow_model(in_data={"input": {}}, model_url=url)


def test_run_kubeflow_model_automatic_vsh_no_request_header_raise_exception():
    with pytest.raises(ValueError):
        utils.run_kubeflow_model(in_data={"input": {}}, model_url=KUBEFLOW_MODEL_URL)


def test_run_kubeflow_model_automatic_vsh_raise_exception_internal_server_error():
    df = TEST_DF.drop(columns="DEPTH")
    in_data = {
        "input": [
            {
                "well": df.well_name.iloc[0],
                "input_logs": df.fillna(-9999).to_dict(orient="list"),
                "keyword_arguments": VSH_KWARGS,
            }
        ]
    }
    with pytest.raises(ValueError):
        utils.run_kubeflow_model(
            in_data=in_data, model_url=KUBEFLOW_MODEL_URL, header=VSH_HEADER
        )


def test_run_deployed_models_call_kubeflow_model_automatic_vsh_return_only_vsh():
    df = TEST_DF.rename(columns={"DENC": "DEN"}).fillna(
        VSH_KWARGS["nan_numerical_value"]
    )
    original_columns = set(df.columns)
    output_df = utils.run_deployed_model(
        df=df,
        id_column=ID_COLUMN,
        keyword_arguments=VSH_KWARGS,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        header=VSH_HEADER,
    )
    output_columns = set(output_df.columns)
    assert output_columns - original_columns == {
        "VSH_AUT",
        "well",
    }, "Curve 'VSH_AUT' was not added to the dataframe"


def test_run_deployed_model_call_kubeflow_model_missing_DEPTH_column_raise_exception():
    df = TEST_DF.drop(columns="DEPTH")
    with pytest.raises(ValueError):
        utils.run_deployed_model(
            df=df,
            id_column=ID_COLUMN,
            keyword_arguments=VSH_KWARGS,
            kubeflow_model_url=KUBEFLOW_MODEL_URL,
            header=VSH_HEADER,
        )


def test_run_deployed_models_call_kubeflow_model_automatic_vsh_return_composite_curves():
    df = TEST_DF.rename(columns={"DENC": "DEN"}).fillna(
        VSH_KWARGS["nan_numerical_value"]
    )
    original_columns = set(df.columns)
    keyword_arguments = VSH_KWARGS.copy()
    keyword_arguments["return_only_vsh_aut"] = False
    output_df = utils.run_deployed_model(
        df=df,
        id_column=ID_COLUMN,
        keyword_arguments=keyword_arguments,
        kubeflow_model_url=KUBEFLOW_MODEL_URL,
        header=VSH_HEADER,
    )
    output_columns = set(output_df.columns)
    assert output_columns - original_columns == {
        "BS_CHANGE",
        "GR_sh",
        "GR_ss",
        "VSH_AUT",
        "VSH_AUT_P025",
        "VSH_AUT_P10",
        "VSH_AUT_P16",
        "VSH_AUT_P84",
        "VSH_AUT_P90",
        "VSH_AUT_P975",
        "VSH_GR_AUT",
        "VSH_GR_AUT_INTER_RATIO",
        "VSH_GR_AUT_QCFLAG",
        "well",
    }, "Curve 'VSH_AUT' and the composite VSH curves were not added to the dataframe"
